import time
from datetime import datetime, timedelta
from tkinter import *
from tkinter.ttk import Progressbar
from typing import  Optional
from cytomat.parameters import Parameters
from cytomat.status import OverviewStatus
from .manager_settings import ManagerSettings
from .db import DB
from .cytomat_plus import CytomatPlus

class GUI:
    __root: Tk
    __cytomat: CytomatPlus
    __DB: DB
    __manager_settings: ManagerSettings
    def __init__(self):
        self.__root = Tk()
        self.__root.title('Cytomat Gui')
        self.__root.geometry('1920x1080')

        self.__parameters = Parameters()
        self.__parameters.load()
        #self.__root.attributes('-fullscreen', True)

        self.__cytomat = CytomatPlus(self.__parameters.COM_port)
        self.__DB = DB()

        self.__manager_settings = ManagerSettings(db = self.__DB, root = self.__root, gui = self)
        self.__manager_settings.create_overview()

        self.__interrupted: bool = False
        self.checker: bool = False
        self.busy_v: bool = False
        self.__pause: bool = False
        self.__root.protocol('WM_DELETE_WINDOW', self.on_close)
        self.busy_row: Optional[int] = None
        #contains all slots where role = 0 in form of a dictionary 
        self.__db_data: list[dict] = self.__DB.get_rows_role_dict(role = 0)
        self.after_id = None
        self.after_id_2 = None
        self.check_if_transfer_occupied()

        self.__list_of_row_objects: Optional[list[GUI_row]] = None
        self.__header_1: Optional[Label] = None
        self.__header_2: Optional[Label] = None
        self.__header_3: Optional[Label] = None
        self.__header_4: Optional[Label] = None
        self.__setting_button: Optional[Button] = None

        self.__ran_all_foc: Optional[bool] = None

        self.__insert_window: Optional[Toplevel] = None
        self.year: Optional[StringVar] = None
        self.month: Optional[StringVar] = None
        self.day: Optional[StringVar] = None
        self.index_: Optional[StringVar] = None

        self.wait_for_submit: Optional[IntVar] = None
        self.submit_insert: Optional[bool] = None

        self.__progress_window: Optional[Toplevel] = None
        self.__progressbar: Optional[Progressbar] = None
        self.__status_label: Optional[Label] = None
        self.__pause_btn: Optional[Button] = None

        self.__reboot_window: Optional[Toplevel] = None
        self.__cyto_init_btn: Optional[Button] = None
        self.__continue_btn: Optional[Button] = None
        self.__get_back_btn: Optional[Button] = None

        self.__interrupt_window_submit: Optional[IntVar] = None #Type later: IntVar
        self.current_task_idx: Optional[int] = None

        self.submit_delete_wait: Optional[IntVar] = None
        self.__submit_delete_window: Optional[Toplevel] = None
        self.__submit_delete_label: Optional[Label] = None
        self.__submit_delete_btn: Optional[Button] = None
        self.__cancel_delete: [Button] = None
        self.submit_delete: Optional[bool] = None

        self.interrupt_state = None
        self.__display_interrupt_action:Optional[Toplevel] = None
        self.__display_action_label: Optional[Label] = None
        self.__header_run: Optional[Label] = None
        self.__run_all_btn: Optional[Button] = None
        self.__run_specific_btn: Optional[Button] = None
        self.__choose_specific_window: Optional[Toplevel] = None
        self.__specific_plates_dict: Optional[dict[int, GUI_row]] = {}
        self.__multiple_media_change_btn: Optional[Button] = None
        self.__multiple_eject_btn: Optional[Button] = None
        self.__multiple_measurement_btn: Optional[Button] = None

    def get_root(self):
        return self.__root

    def set_busy_state(self, busy: bool) -> None:
        """
        sets every row to given busy state

        Parameters
        ----------
        busy: bool
            represents if the GUI/Cytomat is in a busy state
        """
        for row in self.__list_of_row_objects:
            row.busy(busy)

    def get_cytomat_status(self) -> OverviewStatus:
        return self.__cytomat.get_status()

    def create_overview(self) -> None:
        """
        Creates Overview contents and sets arrangement settings
        """
        self.set_dynamic_window_settings()
        self.create_headers()
        self.create_setting_button()
        self.create_row_objects()
        self.create_row_run_all()

    #every row in th gui consists of an object of the gui_row class
    def create_row_objects(self) -> None:
        self.__list_of_row_objects = []
        """
        Creates one row of type gui_row for each row in db_data and adds row object to list
        """
        for row_idx, data_dict in enumerate(self.__db_data):
            row = GUI_row(row_idx+2, root=self.__root, gui = self, db = self.__DB, cytomat=self.__cytomat, transfer_oc=self.checker, data = data_dict)
            self.__list_of_row_objects.append(row)

    def update_overview(self) -> None:
        """
        each row ist getting flashed with updated db_data content
        """
        self.__db_data = self.__DB.get_rows_role_dict()
        self.create_row_objects()

    def set_dynamic_window_settings(self) -> None:
        """
        Overview grid is set proportional to window size
        """
        self.__root.columnconfigure(0, weight=1)
        self.__root.columnconfigure(1, weight=1)
        self.__root.columnconfigure(2, weight=1)
        self.__root.columnconfigure(3, weight=1)
        self.__root.columnconfigure(4, weight=1)
        self.__root.columnconfigure(5, weight=1)

        for i in range(1, len(self.__db_data) + 2):
            self.__root.rowconfigure(i, weight = 1)


    def create_headers(self) -> None:
        """
        Defines and places the headers
        """
        y = 30
        self.__header_1: Label = Label(self.__root, text='SlotNumer', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_1.grid(column=0, row=0, pady = 20, sticky = 'ew', columnspan= 1, ipadx = len('SlotNumber'), ipady=y)

        self.__header_2: Label = Label(self.__root, text='Role', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_2.grid(column=1, row=0, pady = 20, sticky = 'ew', columnspan= 1, ipadx = len('Role'), ipady=y)

        self.__header_3: Label = Label(self.__root, text='AssignedPlatePlateId', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5 , font=('Helvetica', 16))
        self.__header_3.grid(column=2, row=0, pady = 20, sticky = 'new', columnspan= 1, ipadx= len('AssignedPlatePlateId'), ipady=y)

        self.__header_4: Label = Label(self.__root, text='command', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_4.grid(column=3, row=0, pady = 20, sticky = 'new', columnspan= 3, ipadx = len('commands'), ipady=y)

    def create_setting_button(self):
        """
        Defines and places the setting Button
        """
        self.__setting_button = Button(self.__root, text = "Settings", command = self.setting_button_callback)
        self.__setting_button.grid(row = 0, column = 5, ipadx = 15, ipady = 15, sticky='ne', pady=5)

    def setting_button_callback(self):
        self.__manager_settings.display_window()

    def create_row_run_all(self)->None:
        """
        creates a row to control commands of all plates at once

        """
        self.__row_one_frame = Frame(self.__root)
        self.__row_one_frame.grid(row = 1, column=0, columnspan=3, sticky='nsew')

        self.__row_one_frame.columnconfigure(0, weight=1)
        self.__row_one_frame.columnconfigure(1, weight=2)
        self.__row_one_frame.columnconfigure(2, weight=2)
        self.__row_one_frame.rowconfigure(0, weight=1)

        self.__header_run = Label(self.__row_one_frame, text='Run:', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5 , font=('Helvetica', 16))
        self.__header_run.grid(row = 0, column = 0, sticky = 'nsew', columnspan= 1, ipadx= len('run'))

        self.__run_all_btn = Button(self.__row_one_frame, text = 'All', command=self.run_all_btn_callback, bg= 'yellow', relief='raised', borderwidth= 2, pady = 5 , font=('Helvetica', 16))
        self.__run_all_btn.grid(row = 0, column = 1, sticky = 'nsew', columnspan= 1, ipadx= len('all'))

        self.__run_specific_btn = Button(self.__row_one_frame, text='Specific', command= self.run_specific_btn_callback,  bg = 'green',  relief='raised', borderwidth= 2, pady = 5 , font=('Helvetica', 16))
        self.__run_specific_btn.grid(row = 0, column = 2, sticky = 'nsew', columnspan= 1)

        self.__multiple_media_change_btn = Button(self.__root, text ='Media-change', command=self.multiple_media_change_callback, relief='raised', font=('Helvetica', 16))
        self.__multiple_media_change_btn.grid(row = 1, column = 3, sticky ='nsew', columnspan= 1)

        self.__multiple_eject_btn = Button(self.__root, text = 'Eject', command = self.multiple_eject_callback, relief='raised', font=('Helvetica', 16))
        self.__multiple_eject_btn.grid(row = 1, column = 4, sticky = 'nsew', columnspan= 1)

        self.__multiple_measurement_btn = Button(self.__root, text = 'Measurement', command=self.multiple_measurement_callback, relief='raised', font=('Helvetica', 16))
        self.__multiple_measurement_btn.grid(row = 1, column = 5, sticky = 'nsew', columnspan= 1)

    def run_all_btn_callback(self)->None:
        self.__run_all_btn.config(bg="green")
        self.__run_specific_btn.config(bg="yellow")

        self.__specific_plates_dict = self.get_assigned_plate_row_objects()

    def run_specific_btn_callback(self)->None:
        self.create_choose_specific_window()

        self.__run_all_btn.config(bg="yellow")
        self.__run_specific_btn.config(bg="green")

    def create_choose_specific_window(self)->None:
        """
        creates a window of 5 x n Buttons which represent all usable slots. It allows to mark all slots separately to use at once
        """
        self.__choose_specific_window = Toplevel(self.__root)
        self.__choose_specific_window.title("Choose Plates")
        self.__choose_specific_window.geometry('400x400')

        assigned_plate_row_objects_list: list[GUI_row] = self.get_assigned_plate_row_objects()

        for idx, row in enumerate(assigned_plate_row_objects_list):
            column = idx// 5
            row_idx = idx % 5
            if row.slot in self.__specific_plates_dict.keys():
                toggle_state = BooleanVar(value = True)
                bg = 'green'
            else:
                toggle_state = BooleanVar(value = False)
                bg = 'red'
            square_button = Button(self.__choose_specific_window, text= row.slot, bg = bg)
            square_button.config(command=lambda b = square_button, toggle = toggle_state, row_obj = row: self.square_button_callback(b, toggle, row_obj))

            square_button.grid(row = row_idx, column = column, sticky='nsew')
        for i in range(len(self.__list_of_row_objects)//5 +1):
            self.__choose_specific_window.columnconfigure(i, weight = 1)

        self.__choose_specific_window.rowconfigure(0, weight = 1)
        self.__choose_specific_window.rowconfigure(1, weight=1)
        self.__choose_specific_window.rowconfigure(2, weight=1)
        self.__choose_specific_window.rowconfigure(3, weight=1)
        self.__choose_specific_window.rowconfigure(4, weight=1)

    def get_assigned_plate_row_objects(self)->list:
        """
        returns a list of all row objects which have assigned plates
        """
        assigned_plate_row_objects_list = []
        for row in self.__list_of_row_objects:
            if row.plate:
                assigned_plate_row_objects_list.append(row)
        return assigned_plate_row_objects_list

    def square_button_callback(self, btn: Button, tog: BooleanVar, row_obj)->None:
        """
        updates the self.__choose_specific_window widget, and self.__specific_slots_list which is used to run multiple plate commands at 'once'
        Parameters
        ----------
        btn : Button
            object of caller Button, used to update the button color
        tog: BooleanVar
            toggle value: should the slot be used?
        row_obj: GUI_row
            row obj used to update self.__specific_slots_list
        """
        current = tog.get()
        tog.set(not current)

        if tog.get():
            btn.config(bg="green")
            self.__specific_plates_dict[row_obj.slot] = row_obj
        else:
            btn.config(bg='red')
            del self.__specific_plates_dict[row_obj.slot]

    def multiple_media_change_callback(self):
        for key_row in self.__specific_plates_dict.keys():
            self.__specific_plates_dict[key_row].callback_mediachange()

    def multiple_measurement_callback(self):
        for key_row in self.__specific_plates_dict.keys():
            self.__specific_plates_dict[key_row].callback_measurement()

    def multiple_eject_callback(self)->None:
        """
        loops through all row obj in self.__specific_plates_dict and ejects the associated plate. If the Plate has not been
        lifted from the transfer station 10 sec. after it has been ejected it gets inserted again.

        """
        for key_row in self.__specific_plates_dict.keys():

            self.__specific_plates_dict[key_row].callback_eject()
            time_start = time.time()
            max_time = 10
            bool_insert = True
            while time.time() - time_start < max_time:
                if self.get_cytomat_status().transfer_station_occupied:
                    time.sleep(0.5)
                else:
                    bool_insert = False
                    break
            if bool_insert:
                self.__specific_plates_dict[key_row].callback_insert()

    """
    """
    def set_run_all_foc_at_time(self, bool_run: bool = False, timer_hour: int = 3, timer_minute: int = 0, time_buffer: int = 5) -> None:
        """
        controls the check_time_run_all_foc to run the run_all_foc at a specific time. Gets called in ManagerSettings class

        Parameters
        ----------
        bool_run: bool
            should the run all foc run

        timer_hour: int
            at what hour should run_all_foc run

        timer_minute: int
            at what minute should run_all_foc run

        time_buffer: int
            how many minutes of buffer should run_all_foc have to run(default 5 min)
        """
        if bool_run is True:
            self.__ran_all_foc = False
            self.check_time_run_all_foc(timer_hour=timer_hour, timer_minute=timer_minute, time_buffer=time_buffer)

        elif self.after_id_2 is not None:
            self.__root.after_cancel(self.after_id_2)

    def check_time_run_all_foc(self, timer_hour, timer_minute, time_buffer: int = 5) -> None:
        """
        checks if current time is in the interval of given time plus buffer, if so run_all_foc is called else it recalls itself with a delay of 10000ms

        Parameters
        ----------
         timer_hour: int
            at what hour should run_all_foc run

        timer_minute: int
            at what minute should run_all_foc run

        time_buffer: int
            how many minutes of buffer should run_all_foc have to run(default 5 min)
        """
        if not self.__root.winfo_exists():
            return

        time_now = datetime.now()

        midnight = datetime(time_now.year, time_now.month, time_now.day)
        timer = midnight + timedelta(hours=timer_hour, minutes=timer_minute)

        time_buffer_delta = timedelta(minutes=time_buffer)

        if timer <= time_now <= timer + time_buffer_delta:
            if self.__ran_all_foc is False:
                self.run_all_foc()
                self.__ran_all_foc = True

        if time_now > timer + time_buffer_delta:
            self.__ran_all_foc = False

        self.after_id_2 = self.__root.after(10000, self.check_time_run_all_foc, timer_hour, timer_minute, time_buffer)

    def run_all_foc(self) -> None:
        """
        loops through every row in Database and lets the task_manager execute measurement. In the meantime the GUI is blocked/busy state True
        """
        self.set_busy_state(True)
        data_rows: list[dict] = self.__DB.get_rows_role_dict(role = 0)

        for data in data_rows:
            if data['plate'] is None:
                continue
            slot = data['slot']
            plate = data['plate']
            print(f"slot: {slot}, plate: {plate}")
            try:
                self.task_manager(self.__cytomat.plate_handler.get_plate_do_measurement_bring_back_plate(slot_a = slot, plate = plate))

            except Exception as e:
                print(f'execute canceled due to: {e}')

        self.set_busy_state(False)
        self.busy_v = False
        self.update_overview()

    def check_if_transfer_occupied(self) -> None:
        """
        every 100ms the function calls itself as long the main window exists(without blocking the main thread) and checks if the transfer station state has changed.
        If so the function updates the overview
        """
        if not self.__root.winfo_exists():
            return

        transfer_station_occupied = self.get_cytomat_status().transfer_station_occupied
        if self.checker != transfer_station_occupied:
            self.checker = transfer_station_occupied
            self.update_overview()

        self.after_id = self.__root.after(25, self.check_if_transfer_occupied)

    def on_close(self) -> None:
        """
        Kills all the background operations to close the main window properly without raising an error
        """
        if self.after_id is not None:
            self.__root.after_cancel(self.after_id)

        if self.after_id_2 is not None:
            self.__root.after_cancel(self.after_id_2)
        self.__root.destroy()

    def create_insert_new_plate_window(self, slot) -> None:
        """
        creates a window to insert new plates. It consists of a year, month, day and index Entry, which can be in- or decresed by a plus or minus button.

        Parameters
        ----------
        slot: int
            slot number to insert new plate
        """
        self.__insert_window = Toplevel(master=self.__root)
        self.__insert_window.title('insert window')
        self.__insert_window.geometry('1000x400+400+400')

        self.__insert_window.columnconfigure(0, weight=1)
        self.__insert_window.columnconfigure(1, weight=1)
        self.__insert_window.columnconfigure(2, weight=1)
        self.__insert_window.columnconfigure(3, weight=1)
        self.__insert_window.columnconfigure(4, weight=1)
        self.__insert_window.columnconfigure(5, weight=1)

        self.__insert_window.rowconfigure(0, weight=1)
        self.__insert_window.rowconfigure(1, weight=1)
        self.__insert_window.rowconfigure(2, weight=1)
        self.__insert_window.rowconfigure(2, weight=1)

        current_year = datetime.now().year
        current_month = datetime.now().month
        current_day = datetime.now().day

        self.year = StringVar()
        self.month = StringVar()
        self.day = StringVar()
        self.index_ = StringVar()

        self.year.set(f'{current_year:02}')
        self.month.set(f'{current_month:02}')
        self.day.set(f'{current_day:02}')
        self.index_.set('a')

        year_label = Label(master = self.__insert_window, text= 'year', font = ('Helvetica', 14), relief='solid')
        month_label = Label(master = self.__insert_window, text = 'month', font = ('Helvetica', 14),relief='solid')
        day_label = Label(master = self.__insert_window, text = 'day', font = ('Helvetica', 14), relief='solid')
        index_label = Label(master = self.__insert_window, text ='index', font = ('Helvetica', 14), relief='solid')

        year_label.grid(row = 1, column = 0, ipady = 20, sticky = 'ew')
        month_label.grid(row = 1, column = 1, ipady = 20, sticky = 'ew')
        day_label.grid(row = 1, column = 2, ipady = 20, sticky = 'ew')
        index_label.grid(row = 1, column = 3, ipady = 20, sticky = 'ew')

        entry_year = Entry(master = self.__insert_window, textvariable = self.year,  state = 'readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_month = Entry(master =self.__insert_window, textvariable =self.month,  state='readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_day = Entry(master =self.__insert_window, textvariable = self.day, state='readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_idx = Entry(master =self.__insert_window,textvariable = self.index_, state = 'readonly',relief = 'solid', font = ('Helvetica', 14))

        entry_year.grid(row = 2, column = 0, ipady = 40, ipadx = 80)
        entry_month.grid(row = 2, column = 1,ipady = 40, ipadx = 80)
        entry_day.grid(row = 2, column = 2, ipady = 40, ipadx = 80)
        entry_idx.grid(row = 2, column = 3,ipady = 40, ipadx= 80)

        submit_btn = Button(master =self.__insert_window, text = "Submit",command = lambda: self.submit_btn_callback(slot), font = ('Helvetica', 12), relief = 'solid')
        submit_btn.grid(row=0, column=4, ipadx = 80, ipady = 40, rowspan=4, sticky='ns')

        year_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.year_change(0), font = ('Helvetica', 20), relief = 'solid')
        year_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.year_change(1), font = ('Helvetica', 20), relief = 'solid')
        year_btn_plus.grid(row = 0, column=0, sticky = 'ew', ipady = 40)
        year_btn_minus.grid(row = 3, column=0, sticky = 'ew', ipady = 40)

        month_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.month_change(0), font = ('Helvetica', 20), relief = 'solid')
        month_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.month_change(1), font = ('Helvetica', 20), relief = 'solid')
        month_btn_plus.grid(row = 0, column = 1, sticky = 'ew', ipady = 40)
        month_btn_minus.grid(row = 3, column = 1, sticky = 'ew', ipady = 40)

        day_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.day_change(0), font = ('Helvetica', 20), relief = 'solid')
        day_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.day_change(1), font = ('Helvetica', 20), relief = 'solid')
        day_btn_plus.grid(row = 0, column = 2, sticky = 'ew', ipady = 40)
        day_btn_minus.grid(row = 3, column = 2, sticky = 'ew', ipady = 40)

        idx_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.idx_change(0), font = ('Helvetica', 20), relief = 'solid')
        idx_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.idx_change(1), font = ('Helvetica', 20), relief = 'solid')
        idx_btn_plus.grid(row = 0, column = 3, sticky = 'ew', ipady = 40)
        idx_btn_minus.grid(row=3, column = 3, sticky = 'ew', ipady = 40)

        self.__insert_window.protocol('WM_DELETE_WINDOW',self.on_close_insert_window)

        self.__insert_window.update()
        self.__insert_window.update_idletasks()
        self.wait_for_submit = IntVar()

    def on_close_insert_window(self) -> None:
        """
        gets called as soon the insert window is closed and
        """
        self.submit_insert = False
        self.close_insert_window()
        #Boolean variable which is used to continue inside callback(gui_row)
        self.wait_for_submit.set(1)

    def close_insert_window(self) -> None:
        self.__insert_window.destroy()

    def submit_btn_callback(self, slot: int) -> None:
        """
        callback function to the submit button, assigns new plate to a given slot in Database and sets flags for completion
        of insert and for success of insert

        Parameters
        ----------
        slot: int
            the slot number whew the new plate should be assigned to
        """
        plate_id = self.year.get() + self.month.get() + self.day.get() + self.index_.get()

        if self.__DB.is_plate_assigned(plate_id):
            print("This Plate is already assigned to a Slot")
            return
        self.__DB.assigne_plate(plate_id=plate_id, slot = slot)
        self.close_insert_window()
        self.update_overview()
        self.submit_insert = True
        self.wait_for_submit.set(1)

    def year_change(self, p_o_m: int) -> None:
        """
        changes the current year in entry_year of the inset window by +-1

        Parameters
        ----------
        p_o_m: int
            stands for plus or minus, 0 is plus, 1 is minus
        """
        curr_year = int(self.year.get())
        if p_o_m == 0:
            new_year = curr_year + 1
        else:
            new_year = curr_year - 1

        self.year.set(f'{new_year:02}')

    def month_change(self, p_o_m: int) -> None:
        """
        changes the month year in entry_month of the inset window by +-1
        lowest is 1, highest 12

        Parameters
        ----------
        p_o_m: int
            stands for plus or minus, 0 is plus, 1 is minus
        """
        curr_month = int(self.month.get())
        if p_o_m == 0:
            new_month = curr_month +1
        else:
            new_month = curr_month -1

        if new_month > 12:
            new_month = 1

        if new_month < 1:
            new_month = 12
        self.month.set(f'{new_month:02}')

    def day_change(self, p_o_m: int) -> None:
        """
        changes the current day in entry_day of the inset window by +-1
        lowest is 1, highest 31

        Parameters
        ----------
        p_o_m: int
            stands for plus or minus, 0 is plus, 1 is minus
        """
        curr_day = int(self.day.get())
        if p_o_m == 0:
            new_day = curr_day +1
        else:
            new_day = curr_day -1

        if new_day >31:
            new_day = 1

        if new_day <1:
            new_day =31

        self.day.set(f'{new_day:02}')

    def idx_change(self, p_o_m: int) -> None:
        """
        changes the current letter in entry_idx of the inset window by +-1(alphabetic order)
        lowest is a, highest z

        Parameters
        ----------
        p_o_m: int
            stands for plus or minus, 0 is plus, 1 is minus
        """
        curr_index = self.index_.get()
        curr_ascii = ord(curr_index)
        if p_o_m == 0:
            new_ascii = curr_ascii + 1
        else:
            new_ascii = curr_ascii - 1

        if new_ascii > ord('z'):
            new_index = 'a'
        elif new_ascii < ord('a'):
            new_index = 'z'
        else:
            new_index = chr(new_ascii)

        self.index_.set(new_index)

    def create_progress_window(self) -> None:
        """
        Creates a progress window with opens when a command has been clicked. Consisting of a progress bar, status label and an interrupt button
        """
        self.__progress_window = Toplevel(master=self.__root)
        self.__progress_window.geometry('800x200+400+400')
        self.__progress_window.title('Progress window')
        #self.__progress_window.overrideredirect(True)

        self.__progressbar = Progressbar(self.__progress_window, maximum=100)
        self.__progressbar.grid(row = 0, column=0, columnspan=2, padx=5, pady = 50, sticky = 'new', ipadx = 200)

        self.__status_label = Label(self.__progress_window, text = "status")
        self.__status_label.grid(row = 0, column=1, columnspan=2, ipadx = 200, pady = 50, sticky = 'sew')

        self.__pause_btn = Button(self.__progress_window, text="Pause", bg = 'red', command = self.pause_btn_callback)
        self.__pause_btn.grid(row = 0, column = 3, ipadx = 100, ipady = 50, pady = 20, padx=5)

        self.__progress_window.update()

    def close_progress_window(self) -> None:
        self.__progress_window.destroy()

    def pause_btn_callback(self) -> None:
        self.__pause = True
        self.create_interrupt_window()
        print("set .__pause to True")

    def get_interrupted(self):
        return self.__interrupted

    def set_interrupted(self, vbool: bool) -> None:
        self.__interrupted = vbool

    def update_progressbar(self, percent: float) -> None:
        self.__progressbar['value'] = percent
        self.__root.update_idletasks()

    def update_status_label(self, text: str) -> None:
        self.__status_label.config(text = text)
        self.__root.update()

    def is_progress_window_exists(self)->bool:
        try:
            return self.__progress_window.winfo_exists()

        except Exception:
            return False

    def task_manager(self, task_list)->None:
        """
        Gets a list of commands and executes them in cytomat.execute() one by one while checking if no error or pause has occurred.
        Parameters
        ----------
        task_list : list[Callable[[], PlateShuttleSystemStatus]]
        """
        print("task manager")
        task_completed = None
        for idx, command in enumerate(task_list):
            self.current_task_idx = idx
            if self.__pause is True:
                self.__pause = False
                self.__root.wait_variable(self.__interrupt_window_submit)

                if self.interrupt_state != 'continue':
                    task_completed = False
                    print(task_completed)
                    return

            try:
                self.__cytomat.execute(command, timeout=120)

            except Exception as error:
                print(f"error: {error}")
                self.error_handler(error)
                return

            if self.is_progress_window_exists():
                self.update_progressbar(idx/len(task_list) * 100)
                self.update_status_label(command.__name__)
                self.__progress_window.update()
            time.sleep(0.5)
        print("done id: task_manager")
        task_completed = True
        print(task_completed)
        return

    def error_handler(self, error) -> None:
        print(error)
        #TODO: Error handling

    def create_interrupt_window(self) -> None:
        """
        Creates a window to Initialize the Plate handler, gets called after a command has been interrupted
        """
        self.__reboot_window = Toplevel(master=self.__root)
        self.__reboot_window.title('Interrupt window')
        self.__reboot_window.geometry('600x350+400+400')
        self.__reboot_window.overrideredirect(True)

        self.__interrupt_window_submit = IntVar()

        self.__cyto_init_btn = Button(self.__reboot_window, bg = 'red', text="Re-Init Cytomat PH", command = self.callback_cyto_init_btn, relief='solid', font=('Helvetica', 16))
        self.__cyto_init_btn.pack(expand=True, fill='both')

        self.__continue_btn = Button(self.__reboot_window, text="Continue", command = self.callback_continue_btn, font =('Helvetica', 16))
        self.__continue_btn.pack(expand=True, fill='both')

        self.__get_back_btn = Button(self.__reboot_window, text = "Undo command", command=lambda: self.callback_get_back_btn(), font =('Helvetica', 16))
        self.__get_back_btn.pack(expand = True, fill = 'both')

        self.__reboot_window.update()
        self.__reboot_window.update_idletasks()

    def callback_get_back_btn(self) -> None:
        self.interrupt_state = ['get_back', self.current_task_idx]
        self.__interrupted = True
        self.__reboot_window.destroy()
        self.__interrupt_window_submit.set(1)

    def callback_cyto_init_btn(self):
        self.interrupt_state = ['reinitialize']
        self.__interrupted = True
        self.__reboot_window.destroy()
        self.__interrupt_window_submit.set(1)

    def callback_continue_btn(self):
        self.interrupt_state = 'continue'
        self.__interrupt_window_submit.set(1)
        self.__reboot_window.destroy()
    """
    """
    def create_interrupt_actions_window(self):
        self.__display_interrupt_action = Toplevel(master =self.__root)
        self.__display_interrupt_action.geometry('600x350+400+400')
        self.__display_interrupt_action.title('Interrupt Action')

        self.__display_action_label = Label(master=self.__display_interrupt_action, text ="currently doing:")
        self.__display_action_label.pack(expand=True, fill='both')

    def modify_display_label(self, text_):
        self.__display_action_label.config(text=text_)
        self.__display_interrupt_action.update()

    def destroy_display_interrupt(self):
        self.__display_interrupt_action.destroy()

    def create_submit_delete_window(self, plate: str):
        self.submit_delete_wait = IntVar()

        self.__submit_delete_window = Toplevel(master=self.__root)
        self.__submit_delete_window.geometry('600x350+400+400')
        self.__submit_delete_window.title('Submit Delete')
        self.__submit_delete_window.overrideredirect(True)

        self.__submit_delete_window.columnconfigure(0, weight=1)
        self.__submit_delete_window.columnconfigure(1, weight=1)
        self.__submit_delete_window.columnconfigure(2, weight=1)
        self.__submit_delete_window.rowconfigure(0, weight=1)
        self.__submit_delete_window.rowconfigure(1, weight=1)
        self.__submit_delete_window.rowconfigure(2, weight=1)

        self.__submit_delete_label = Label(master= self.__submit_delete_window, relief = 'solid', text = f"To delete plate: {plate} press button", font = ('Helvetica', 14), bg = 'lightgrey')
        self.__submit_delete_label.grid(row =0, column=0, sticky='nsew', columnspan=3)

        self.__submit_delete_btn = Button(master = self.__submit_delete_window, command=self.submit_delete_callback, text ="Commit Delete", bg = 'red', font = ('Helvetica', 14))
        self.__submit_delete_btn.grid(row=1, column=1, sticky='nsew')

        self.__cancel_delete = Button(master = self.__submit_delete_window, command=self.cancel_delete, text = "Cancel", font = ('Helvetica', 14))
        self.__cancel_delete.grid(row=2, column=1, sticky='nsew')

    def submit_delete_callback(self):
        self.submit_delete = True
        self.submit_delete_wait.set(1)
        self.__submit_delete_window.destroy()

    def cancel_delete(self):
        self.submit_delete = False
        self.submit_delete_wait.set(1)
        self.__submit_delete_window.destroy()

class GUI_row:
    __label_slot    : Label
    __label_role    : Label
    __label_plate   : Label
    __btn_media     : Button
    __btn_eject     : Button
    __btn_measurement   : Button
    __DB            : DB
    cy              : CytomatPlus
    row_idx         : int
    command_map: dict
    #db_row: DB_row

    def __init__(self, row_idx, root, gui, db, cytomat, transfer_oc: bool, data: dict):
        """
        Every row of the gui consists of an object of the GUI_row class, 6 columns of which 3 are data and 3 commands
        """
        self.__root: Tk = root
        self.cy: CytomatPlus = cytomat
        self.gui: GUI = gui
        self.__DB = db

        self.row_idx = row_idx

        self.slot = data['slot']
        self.role = data['role']
        self.plate = data['plate']
        self.occupied = data['occupied']

        self.transfer_oc = transfer_oc
        self.cyto_status = self.gui.get_cytomat_status()
        self.command_map = self.command_map_()
        self.create()

    def create(self):
        if self.plate is None:
            plate_bool = False
        else: 
            plate_bool = True

        if self.occupied == 0:
            occupied_bool = False
        else:
            occupied_bool = True

        command_map = self.command_map[(self.transfer_oc, plate_bool, occupied_bool)]
        
        cmd_left =      command_map['cmd_left']
        bg_left =       command_map['color_left']
        text_left =     command_map['text_left']
        state_left =    command_map['state_left']

        cmd_middle =    command_map['cmd_middle']
        bg_middle =     command_map['color_middle']
        text_middle =   command_map['text_middle']
        state_middle =  command_map['state_middle']

        cmd_right =     command_map['cmd_right']
        bg_right =      command_map['color_right']
        text_right =    command_map['text_right']
        state_right =   command_map['state_right']
        
        self.__btn_media = Button(self.__root, text = text_left, bg = bg_left, command= cmd_left, state = state_left, font=('Helvetica', 14))
        self.__btn_eject = Button(self.__root, text = text_middle, bg = bg_middle, command = cmd_middle, state = state_middle, font=('Helvetica', 14))
        self.__btn_measurement = Button(self.__root, text = text_right, bg = bg_right, command = cmd_right, state = state_right, font=('Helvetica', 14))

        self.__label_slot =Label(self.__root, text=self.slot , relief='solid',  pady = 1, padx = 1, font=('Helvetica', 14))
        self.__label_role =Label(self.__root, text=self.role , relief='solid',  pady = 1, padx = 1, font=('Helvetica', 14))
        self.__label_plate=Label(self.__root, text=self.plate, relief='solid', pady = 1, padx = 1, font=('Helvetica', 14))

        if self.gui.busy_v is True:
            self.busy(True)
            
        if self.gui.busy_row == self.row_idx:
            self.busy_row(True)

        if self.role == 1:
            self.busy(True)

        self.__label_slot.grid(  column = 0, row = self.row_idx, sticky='nsew', pady = 1)
        self.__label_role.grid(  column = 1, row = self.row_idx, sticky='nsew', pady = 1)
        self.__label_plate.grid( column = 2, row = self.row_idx, sticky='nsew', pady = 1)
        self.__btn_media.grid(   column=  3, row = self.row_idx, sticky = 'nsew', ipadx = len('mediachange'))
        self.__btn_eject.grid(   column=  4, row = self.row_idx, sticky = 'nsew', ipadx = len('Insert'))
        self.__btn_measurement.grid(column=  5, row = self.row_idx, sticky ='nsew', ipadx = len('Test'))

    def command_map_(self)->dict:
        """
        Maps all possible Button-states according to three bools: 1. transfer_station_occupied? 2. plate_is_assigned? 3. is_slot_occupied?

        """
        cmd_map = {
            (False, False, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_register(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Register',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (False, False, True): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'grey', 'text_left' :'None',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'None',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'grey', 'text_right': 'None'},

            (False, True, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_delete(), 'state_middle': 'normal', 'color_middle': 'yellow', 'text_middle':'Delete',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (False, True, True): {'cmd_left': lambda: self.callback_mediachange(), 'state_left': 'normal', 'color_left': 'green', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_eject(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Eject',
                                    'cmd_right': lambda: self.callback_measurement(), 'state_right': 'normal', 'color_right': 'green', 'text_right': 'FOC48'},

            (True, False, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_register(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Register',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (True, False, True): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'grey', 'text_left' :'None',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'None',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'grey', 'text_right': 'None'},

            (True, True, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_insert(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Insert',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (True, True, True): {'cmd_left': lambda: self.callback_mediachange(), 'state_left': 'normal', 'color_left': 'green', 'text_left' :'mediachange',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'Insert',
                                    'cmd_right': lambda: self.callback_measurement(), 'state_right': 'normal', 'color_right': 'green', 'text_right': 'FOC48'},
            }
        return cmd_map
    
    def busy(self, lock: bool) -> None:
        """
        Sets the row to busy state

        Parameters
        ----------

        lock: bool
            parameter for busy state
        """
        if lock is True:
            self.__label_slot.config( bg = 'grey')        
            self.__label_role.config( bg = 'grey')        
            self.__label_plate.config(bg = 'grey')        
            self.__btn_media.config(state = DISABLED, bg = 'grey')
            self.__btn_eject.config(state = DISABLED, bg = 'grey')
            self.__btn_measurement.config(state = DISABLED, bg ='grey')
        else:
            self.__label_slot.config(bg = 'white')      
            self.__label_role.config(bg = 'white')        
            self.__label_plate.config(bg = 'white')        
            self.__btn_media.config(state = NORMAL)
            self.__btn_eject.config(state = NORMAL)
            self.__btn_measurement.config(state = NORMAL)
        self.__root.update()

    def busy_row(self, lock: bool) -> None:
        """
        is a command executed the row is set to busy state/specific color to determine which row is getting executed.
        """
        if lock is True:
            self.__label_slot.config(bg= 'green')
            self.__label_role.config(bg= 'green')
            self.__label_plate.config(bg= 'green') 
        else:
            self.__label_slot.config(bg= 'white')
            self.__label_role.config(bg= 'white')
            self.__label_plate.config(bg= 'white')
        self.__root.update()

    """
    Callback functions of the Buttons
    """
    def callback_register(self) -> None:
        """
        checks if the busy bool is not busy, the assigned role is the right one and if the plate is not assigned already.
        Then sets the Gui to busy and then calls create_insert_new_plate window out of gui
        """
        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storage")
            return

        if self.plate is not None:
            print("The Slot is already assigned to a Plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_insert_new_plate_window(self.slot)
        self.__root.wait_variable(self.gui.wait_for_submit)
        
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def callback_delete(self) -> None:
        """
        Delete logics, sets gui to busy, opens delete window, deletes if submit_delete out of gui is True, then updates overview
        """
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)

        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_submit_delete_window(plate = self.plate)
        self.__root.wait_variable(self.gui.submit_delete_wait)

        if self.gui.submit_delete is True:
            self.delete_plate()

        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()


    def delete_plate(self):
        self.__DB.delete_plate(self.slot)
        print('deleted')

    """
    insert logics
    """
    def callback_insert(self):
        status = self.gui.get_cytomat_status()

        #Basic checks, looking if the Plate is in the right state:
        if status.transfer_station_occupied is False:
            print("There is no plate on the transfer station")
            return

        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storage")
            return

        #Blocking all the other buttons, Gui is busy now
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)

        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager([lambda: self.cy.plate_handler.move_plate_from_transfer_station_to_slot(self.slot)])

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            self.gui.set_interrupted(False)
            self.gui.create_interrupt_actions_window()

            self.choose_interrupt_action(key = 'insert')
        else:
            self.__DB.update_occupied(self.slot, 1)

        self.__root.update()
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def insert(self):
        self.cy.execute([lambda: self.cy.plate_handler.move_plate_from_transfer_station_to_slot(self.slot)], timeout = 20)
 
    def invert_insert_list(self, current_idx: int)->list:
        inverted_list = self.cy.plate_handler.inverted_move_plate_from_transfer_station_to_slot_v2(self.slot)
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list
    """
    eject logics
    """
    def callback_eject(self):
        status = self.gui.get_cytomat_status()

        if status.transfer_station_occupied is True:
            print("The transfer station is occupied")
            return

        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storage")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager(task_list = [lambda: self.cy.plate_handler.move_plate_from_slot_to_transfer_station(self.slot)])

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            self.gui.create_interrupt_actions_window()

            self.gui.set_interrupted(False)

            self.choose_interrupt_action(key = 'eject')
            self.gui.destroy_display_interrupt()

        else:
            self.__DB.update_occupied(self.slot, 0)    #TODO update in re_init_window 

        self.__root.update()
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def eject(self):
        print("running eject")
        self.cy.execute([lambda: self.cy.plate_handler.move_plate_from_slot_to_transfer_station(self.slot)], timeout = 20)
        print("eject Done")

    def invert_eject_list(self, current_idx: int)->list:
        inverted_list = self.cy.plate_handler.inverted_move_plate_from_slot_to_transfer_station_v2(self.slot)
        modified_invert_list = []
        for i in range(current_idx+1):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    """
    mediachange logics
    """
    def callback_mediachange(self):
        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("his Plate is not for mediachange")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager(task_list = self.cy.plate_handler.test3())
        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            """if self.gui.error_raised:
                print("Error raised")"""

            self.gui.set_interrupted(False)
            self.gui.create_interrupt_actions_window()

            self.choose_interrupt_action(key='mediachange')
            self.gui.destroy_display_interrupt()

        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def invert_mediachange_list(self, current_idx):
        inverted_list = self.cy.plate_handler.inverted_test3()
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    """
    measurement logics
    """
    def callback_measurement(self):
        #self.__DB.delete_plate(row_idx=self.row_idx-1)
        if self.gui.busy_v is True:
            print("Cyto is busy")
            return

        if self.role != 0:
            print("This Plate is not for mediachange")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        if self.cyto_status.shovel_occupied is True:
            print("shovel occupied")
            return
        
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()
        self.gui.task_manager(self.measurement())

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            
            self.gui.set_interrupted(False)

            self.gui.create_interrupt_actions_window()
            self.choose_interrupt_action(key='measurement')
            self.gui.destroy_display_interrupt()
                
            """self.__DB.write_log(time_stamp = datetime.now(), action= '', plate_id=self.plate)"""
  
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def measurement(self):
        return self.cy.plate_handler.get_plate_do_measurement_bring_back_plate(plate = self.plate, slot_a = self.slot)

    def invert_measurement_list(self, current_idx):
        inverted_list = self.cy.plate_handler.inverted_get_plate_do_measurement_bring_back_plate(slot_a = self.slot)
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    def init_plate_handler(self):
        self.cy.execute([self.cy.plate_handler.initialize])

    def choose_interrupt_action(self, key):
        command_mapping = {'mediachange': self.invert_mediachange_list,
                           'eject': self.invert_eject_list,
                           'insert': self.invert_insert_list,
                           'measurement': self.invert_measurement_list}

        match self.gui.interrupt_state[0]:
                case 'initialize':
                    self.gui.modify_display_label(text_ ="Currently doing: Initializing plate handler")
                    self.gui.task_manager([self.init_plate_handler])
                case 'get_back':
                    self.gui.modify_display_label(text_ ="Currently doing: Reversing the action")
                    modified_invert_list = command_mapping[key](self.gui.interrupt_state[1])
                    self.gui.task_manager(task_list=modified_invert_list)
                case _: 
                    pass


from cytomat.parameters import Parameters


class ConvertSteps:
    parameters = Parameters()
    parameters.load()

    @classmethod
    def mm_to_steps_x(cls, mm: float) -> int:
        return round(cls.parameters.steps_per_mm_x * mm)

    @classmethod
    def steps_to_mm_x(cls, steps: int) -> float:
        return round(1 / (cls.parameters.steps_per_mm_x / steps), 4)

    @classmethod
    def mm_to_steps_h(cls, mm: float) -> int:
        return round(cls.parameters.steps_per_mm_h * mm)

    @classmethod
    def steps_to_mm_h(cls, steps: int) -> float:
        return round(1 / (cls.parameters.steps_per_mm_h / steps), 4)

    @classmethod
    def mm_to_steps_shovel(cls, mm: float) -> int:
        return round(cls.parameters.steps_per_mm_shovel * mm)

    @classmethod
    def steps_to_mm_shovel(cls, steps: int) -> float:
        return round(1 / (cls.parameters.steps_per_mm_shovel / steps), 4)

    @classmethod
    def deg_to_steps_turn(cls, deg: float) -> int:
        return round(cls.parameters.steps_per_deg_turn * deg)

    @classmethod
    def steps_to_deg_turn(cls, steps: int) -> float:
        return round(1 / (cls.parameters.steps_per_deg_turn / steps), 4)




from tkinter import *
from ..db import DB
from typing import Optional, Any

class ManagerSlots():
    __root: Frame
    __db: DB
    def __init__(self, root, db):
        self.__root = root
        self.__db: DB = db
        self.__row_obj_list = []

        # dict: key: tupel(start: int, end: int), value: SlotsRow()
        self.__dict_of_rows: dict = {}
        self.__reserved_slots = self.__db.get_roled_slots()
        self.__all_slots_list: list = self.__db.get_unroled_slots()

        self.__header_slots: Optional[Label] = None
        self.__entry_bottom: Optional[Entry] = None
        self.__entry_top: Optional[Entry] = None
        self.__label_middle: Optional[Label] = None
        self.__entry_upper: Optional[Entry] = None
        self.__button_enter: Optional[Button] = None

    def create_overview(self):
        self.set_grid_settings()
        self.create_header()
        self.create_entry_area()
        self.update_display()
    
    def create_header(self):
        self.__header_slots = Label(self.__root, text = "Set the Slots wich should be used", font = ('Helvetica', 20), pady = 1, padx = 1, relief='solid')
        self.__header_slots.grid(row = 0, column=0, columnspan = 4, sticky='nsew', ipady = 30)
        
    def create_entry_area(self):
        y = 12
        entry_bottom = Variable(value='bottom')
        entry_upper = Variable(value='upper')
        
        self.__entry_bottom = Entry(self.__root, textvariable = entry_bottom, justify = 'center')
        self.__entry_bottom.grid(row = 1, column = 0, sticky = 'nsew', ipady = y)
        
        self.__label_middle = Label(self.__root, text = '-', font=('Helvetica', 14))
        self.__label_middle.grid(row= 1, column = 1, sticky = 'nsew', ipady = y)
        
        self.__entry_upper = Entry(self.__root, textvariable = entry_upper, justify = 'center')
        self.__entry_upper.grid(row = 1, column = 2, sticky = 'nsew', ipady = y)
    
        self.__button_enter = Button(self.__root, text = 'Enter', command = self.enter_button_callback)
        self.__button_enter.grid(row = 1, column = 3, sticky = 'nsew', ipady = y)
    
    def set_grid_settings(self):
        #self.__root.rowconfigure(0, weight=2)
        #self.__root.rowconfigure(1, weight=1)
        self.__root.rowconfigure(2, weight=1)
        self.__root.rowconfigure(3, weight=2)
        self.__root.rowconfigure(4, weight=2)
        self.__root.columnconfigure(0, weight = 2)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.columnconfigure(2, weight = 2)
        self.__root.columnconfigure(3, weight = 1)
        
    def apply_theme_pack(self, theme_pack: dict) -> None:
        """
        applies theme pack by decoding given theme pack dictionary and configure tkinter widgets

        Parameters
        ----------
        theme_pack: dict
            dictionary with
        """
        self.__theme_pack = theme_pack
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        header = theme_pack['header']
        
        self.__root.config(bg = bg)
        self.__header_slots.config(bg=header, fg = fg)
        self.__label_middle.config(bg = bg, fg = fg)
        self.__entry_upper.config(bg = entry, fg = fg)
        self.__entry_bottom.config(bg = entry, fg = fg)
        self.__button_enter.config(bg = button, fg = fg)
        
        if len(self.__dict_of_rows)<1:
            return
        
        for row in self.__dict_of_rows.values():
            row: SlotsRow
            row.aplay_theme_pack(theme_pack=self.__theme_pack)

        
    def enter_button_callback(self):
        entry_bottom_str = self.__entry_bottom.get()
        entry_upper_str = self.__entry_upper.get()
        self.__entry_bottom.delete(0, END)
        self.__entry_upper.delete(0, END)
        
        if not self.is_entry_valid(entry_bottom= entry_bottom_str, entry_upper = entry_upper_str):
            return
        
        entry_bottom = int(entry_bottom_str)
        entry_upper = int(entry_upper_str)

        entry_list = self.get_int_list_from_entry(entry_b = entry_bottom, entry_u = entry_upper)

        new_slots_list = self.get_excluded_list(list_a= entry_list, list_b=self.__all_slots_list)

        self.update_db(new_slots_list)

        #update variables
        self.__reserved_slots = self.__db.get_roled_slots()
        self.__all_slots_list: list = self.__db.get_unroled_slots()

        self.update_display()

    def is_entry_valid(self, entry_bottom: str, entry_upper: str):
        if not entry_bottom.isdigit() or not entry_upper.isdigit():
            return
        
        entry_bottom_int = int(entry_bottom)
        entry_upper_int = int(entry_upper)
        
        if entry_upper_int < entry_bottom_int:
            return
        
        print(entry_bottom_int)
        entry_list = self.get_int_list_from_entry(entry_b = entry_bottom_int, entry_u = entry_upper_int)
        
        #checks if the entry contains roled slots
        if list(set(entry_list) & set(self.__reserved_slots)):
            return
        
        #checks if the entry is completly cotained in allready used slots
        if not list(set(entry_list) - set(self.__all_slots_list)):
            print(list(set(entry_list) - set(self.__all_slots_list)))
            print('contained')
            return
        
        print("entry valid")
        return True

    def update_display(self):
        consecutive_slots = self.__db.get_consecutive_slots()
        if len(self.__dict_of_rows) > 0:
            for row in self.__dict_of_rows.values():
                row.delete_row()
            self.__dict_of_rows = {}
        
        for idx, key in enumerate(consecutive_slots.keys()):
            self.__dict_of_rows[key] = SlotsRow(root= self.__root, bottom = key[0], upper= key[1], row_idx=idx + 2)
            try:
                self.__dict_of_rows[key].aplay_theme_pack(self.__theme_pack)
            except AttributeError:
                pass

    def get_int_list_from_entry(self, entry_b: int, entry_u: int)->list:
        entry_list = []
        for i in range(entry_b, entry_u+1):
            entry_list.append(i)

        return entry_list

    #gets list a and list b and returns the elemnts witch a contain and b not
    def get_excluded_list(self, list_a: list, list_b: list)->list:
        return list(set(list_a) - set(list_b))

    def update_db(self, list_to_insert: list):
        for slot in list_to_insert:
            self.__db.insert_new_row(slot)


class SlotsRow:
    def __init__(self, root: Toplevel, bottom: int, upper, row_idx: int) -> None:
        self.__root = root
        self.__bottom = bottom
        self.__upper = upper
        self.__row_idx = row_idx
        self.create_row()
    
    def create_row(self):
        self.set_grid_settings()
        self.__label_slots = Label(self.__root, text = f"{self.__bottom} - {self.__upper}", relief='solid')
        self.__label_slots.grid(column=0, columnspan= 3, row=self.__row_idx, sticky='nsew')

        self.__button_delete = Button(self.__root, text= 'delete', command = self.delete_button_callback, relief='solid')
        self.__button_delete.grid(column=3, row=self.__row_idx, sticky='nsew')

    def set_grid_settings(self):
        self.__root.columnconfigure(0, weight=2)
        self.__root.columnconfigure(1, weight=1)
        self.__root.rowconfigure(self.__row_idx, weight = 1)

    def delete_button_callback(self):
        pass

    def delete_row(self):
        self.__label_slots.destroy()
        self.__button_delete.destroy()

    def aplay_theme_pack(self, theme_pack):
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        header = theme_pack['header']

        self.__label_slots.config(bg=bg, fg=fg)
        self.__button_delete.config(bg=button, fg = fg)


import subprocess
import time
from typing import  Callable
from cytomat.plate_handler import PlateHandler
from cytomat.status import OverviewStatus, PlateShuttleSystemStatus
from cytomat.serial_port import SerialPort
from cytomat.parameters import Parameters
from .convert_steps import ConvertSteps

class PlateHandlerPlus(PlateHandler):

    def __init__(self, serial_port: SerialPort) -> None:
        self.parameters = Parameters()
        self.parameters.load()
        super().__init__(serial_port)
        self.lid_holder_slot     = self.parameters.lid_holder_slot
        self.pipet_station_slot  = self.parameters.pipet_station_slot
        self.max_steps_shovel    = self.parameters.max_steps_shovel
        self.measurement_slot    = self.parameters.measurement_slot

        self.convert_steps = ConvertSteps()

              
    def move_x_to_slot(self, slot: int) -> PlateShuttleSystemStatus:
        """time
        Move along to the given slot (only moves along the x axis)

        Parameters
        ----------
        slot
            The target slot
        """
        return self._PlateHandler__serial_port.issue_action_command(f"ll:xp {slot:03}")

    def warning_msg(self):
        print(
            """
    WARNING!!! This command does not check if the handler is in a safe position.
    This can cause crashes. Make sure it is safe to run."""
        )
        inp = input(
            """
            To deactivate this warning message and execute the command: press Y
            To just execute the command: press N
            To exit the script: press E"""
        )

        match inp.upper():
            case "Y":
                self.warning = False
            case "N":
                self.warning = True
            case "E":
                exit()
            case _:
                self.warning_msg()

    from typing import Callable

    def run_shovel_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run the shovel to an absolute mm position."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_shovel(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:sa {steps:05}")]

    def run_shovel_in_relative_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run the shovel relatively in mm."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_shovel(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:sr {steps:05}")]

    def run_turn_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run turn to an absolute position in mm (interpreted as degrees)."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.deg_to_steps_turn(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:da {steps:05}")]

    def run_turn_in_relative_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run turn relatively in mm (interpreted as degrees)."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.deg_to_steps_turn(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:dr {steps:05}")]

    def run_height_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run height to an absolute mm position."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_h(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:ha {steps:05}")]

    def run_height_in_relative_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run height relatively in mm."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_h(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:hr {steps:05}")]

    def run_turntable_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run turntable to an absolute mm position (interpreted as degrees)."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.deg_to_steps_turn(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:ka {steps:05}")]

    def run_turntable_in_relative_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run turntable relatively in mm (interpreted as degrees)."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.deg_to_steps_turn(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:kr {steps:05}")]

    def run_x_axis_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run x-axis to an absolute mm position."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_x(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:xa {steps:05}")]

    def run_x_axis_in_relative_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run x-axis relatively in mm."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_x(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:xr {steps:05}")]

    def run_transfer_station_in_absolute_mm(self, mm: float) -> list[Callable[[], PlateShuttleSystemStatus]] | None:
        """Run transfer station to an absolute mm position (interpreted as x-axis movement)."""
        if self.warning:
            self.warning_msg()
            return None
        steps = self.convert_steps.mm_to_steps_x(mm)
        return [lambda: self.__serial_port.issue_action_command(f"sb:ta {steps:05}")]
    
    def leave_lid_at_holder(self) -> list[Callable[[], PlateShuttleSystemStatus]]| None:
        """

        Returns
        -------
        command_list: list of Callable[[], PlateShuttleSystemStatus]
            a list of frozen commands which execute an action on the cytomat plate handler and
            return a status data object

        """
        if self.parameters.lid_holder_slot is None:
            print("Lid holder not initialized, missing arg type int at the initialisation of Object type CytomatPlus")
            return None

        command_list = [
            lambda: self.rotate_handler_to_slot(self.lid_holder_slot),
            lambda: self.move_x_to_slot(self.lid_holder_slot),
            lambda: self.run_height_in_absolute_mm(26),
            lambda: time.sleep(0.2),
            lambda: self.extend_shovel(),
            lambda: self.run_height_in_absolute_mm(14),
            lambda: time.sleep(0.2),
            lambda: self.retract_shovel(),
            lambda: self.reset_handler_position()]
        return command_list

    def get_lid_from_holder(self)-> list[Callable[[], PlateShuttleSystemStatus]] | None:
        if self.parameters.lid_holder_slot is not None:
            command_list = [
                lambda: self.rotate_handler_to_slot(self.lid_holder_slot),
                lambda: self.move_x_to_slot(self.lid_holder_slot),
                lambda: self.run_height_in_absolute_mm(16),
                lambda: self.run_x_axis_in_relative_mm(-2),
                lambda: self.extend_shovel(),
                lambda: self.run_height_in_absolute_mm(22),
                lambda: self.run_x_axis_in_relative_mm(-1),
                lambda: self.run_x_axis_in_relative_mm(1),
                lambda: self.run_shovel_in_relative_mm(-1),
                lambda: self.run_shovel_in_relative_mm(1),
                lambda: self.run_height_in_absolute_mm(23),
                lambda: self.run_x_axis_in_relative_mm(-1),
                lambda: self.run_x_axis_in_relative_mm(1),
                lambda: self.run_shovel_in_relative_mm(-1),
                lambda: self.run_shovel_in_relative_mm(1),
                lambda: self.run_height_in_absolute_mm(24),
                lambda: self.run_x_axis_in_relative_mm(-1),
                lambda: self.run_x_axis_in_relative_mm(1),
                lambda: self.run_shovel_in_relative_mm(-1),
                lambda: self.run_shovel_in_relative_mm(1),
                lambda: self.run_height_in_absolute_mm(24),
                lambda: self.retract_shovel(),
                lambda: self.reset_handler_position()]
            return command_list

    def move_below_slot(self, slot: int)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.rotate_handler_to_slot(slot),
            lambda: self.move_x_to_slot(slot),
            lambda: self.move_handler_below_slot_height(slot)]
        return command_list

    def pipet_station(self, rows: int, distance_mm: float, start_mm: float) -> list[Callable[[], PlateShuttleSystemStatus]]:
        if (rows * distance_mm) + start_mm > CS.steps_to_mm_shovel(self.max_steps_shovel):
            raise ValueError("The parameters are ot compatible with the Plate handler")

        command_list = [
            lambda: self.move_handler_below_slot_height(self.pipet_station_slot),
            lambda: self.run_shovel_in_relative_mm(start_mm)        
        ]
        for i in range(rows):
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_shovel_in_relative_mm(distance_mm))
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_height_in_relative_mm(15))
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_height_in_relative_mm(-15))
            command_list.append(lambda: time.sleep(0.5))
            
        command_list.append(lambda: self.retract_shovel())
        return command_list

    def do_media_change_from_slot(self, slot: int)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list: list[OverviewStatus]
        first_part = [  lambda: self.move_plate_from_slot_to_handler(slot),
                        lambda: self.leave_lid_at_holder()]

        second_part = self.move_below_slot(slot)

        third_part = self.pipet_station()

        fourth_part = [ lambda: self.get_lid_from_holder(),
                        lambda: self.move_plate_from_handler_to_slot(slot)]

        command_list = first_part + second_part + third_part + fourth_part
        return command_list


    #TODO finisch this func
    def inverted_do_mediachange_from_slot(self, slot: int)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.move_plate_from_handler_to_slot(slot)]
        return command_list

    def get_plate_do_measurement_bring_back_plate(self, plate: str, slot_a: int) -> list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(self.measurement_slot),
            lambda: self.run_foc(plate),
            lambda: self.move_plate_from_slot_to_handler(self.measurement_slot),
            lambda: self.move_plate_from_handler_to_slot(slot_a)]

        return command_list

    def inverted_get_plate_do_measurement_bring_back_plate(self, slot_a: int)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(self.measurement_slot),
            lambda: time.sleep(0.1),
            lambda: self.move_plate_from_slot_to_handler(self.measurement_slot),
            lambda: self.move_plate_from_handler_to_slot(slot_a)]
        return command_list

    def move_plate_from_transfer_station_to_slot_v2(self):
        pass

    def inverted_move_plate_from_transfer_station_to_slot_v2(self, slot):
        pass

    def move_plate_from_slot_a_to_slot_b(self, slot_a: int, slot_b:int)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(slot_b)]

        return command_list

    def run_foc(self, plate, _timeout: int =180):
        cmd_l = f'C:\\labhub\\Import\\FOC48.bat {plate}'
        
        try:
            process = subprocess.run(cmd_l, timeout=_timeout, shell = True)
            print("done")

        except subprocess.TimeoutExpired:
            print(" timeout expired")


    def inverted_move_plate_from_slot_to_transfer_station_v2(self, slot):
        pass

    def do_one_uml(self)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
        lambda: self.run_shovel_in_relative_mm(-18),
        lambda: self.run_height_in_relative_mm(8),
        lambda: time.sleep(0.2),
        lambda: self.run_height_in_absolute_mm(0),
        lambda: time.sleep(0.2)]

        return command_list

    #start position is at the bottom right corner(door view)
    def get_in_mediachange_position(self)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.run_height_in_absolute_mm(0),
            lambda: time.sleep(0.5),
            lambda: self.rotate_handler_to_slot(self.pipet_station_slot),
            lambda: self.run_x_axis_in_absolute_mm(95.5),
            lambda: time.sleep(5),
            lambda: self.extend_shovel(),
            lambda: self.run_shovel_in_relative_mm(-1),
            lambda: time.sleep(0.2)]

        return command_list

    def switch_uml_row(self)->list[Callable[[], PlateShuttleSystemStatus]]:
        command_list = [
            lambda: self.run_x_axis_in_relative_mm(-9),
            lambda: time.sleep(0.2),
            lambda: self.extend_shovel(),
            lambda: self.run_shovel_in_relative_mm(-1)]

        return command_list

    def mediachange_test(self)->list[Callable[[], PlateShuttleSystemStatus]]:
        #possition plate below pipet station, first row, leftest uml direktly under needle
        get_in_start_position = [
            lambda: self.run_height_in_absolute_mm(0),
            lambda: time.sleep(0.5),
            lambda: self.rotate_handler_to_slot(self.pipet_station_slot),
            lambda: self.run_x_axis_in_absolute_mm(95.5),
            lambda: time.sleep(5),
            lambda: self.extend_shovel(),
            lambda: self.run_shovel_in_relative_mm(-1),
            lambda: time.sleep(0.2)
        ]

        do_one_uml = [
            lambda: self.run_shovel_in_relative_mm(-18),
            lambda: self.run_height_in_relative_mm(8),
            lambda: time.sleep(0.2),
            lambda: self.run_height_in_absolute_mm(0),
            lambda: time.sleep(0.2)
        ]

        do_one_row = []
        for i in range(6):
            do_one_row = do_one_row + do_one_uml

        switch_row = [
            lambda: self.run_x_axis_in_relative_mm(-9),
            lambda: time.sleep(0.2),
            lambda: self.extend_shovel(),
            lambda: self.run_shovel_in_relative_mm(-2)

        ]
        homing = [lambda: self.retract_shovel()]
        do_all_uml = []
        for i in range(7):
            do_all_uml += do_one_row + switch_row
        do_all_uml += do_one_row

        command_list = get_in_start_position + do_all_uml + homing + self.get_lid_from_holder()
        return command_list


from dataclasses import dataclass
from logging import log
from cytomat.utils import lazy_load_config_file

@dataclass
class Parameters:
    try:
        python_data = lazy_load_config_file()
        if python_data:
            COM_port = python_data["COM_port"]
            steps_per_mm_h = python_data["steps_per_mm_h"]
            max_steps_h = python_data["max_steps_h"]
            steps_per_mm_x = python_data["steps_per_mm_x"]
            max_steps_x = python_data["max_steps_x"]
            steps_per_mm_shovel = python_data["steps_per_mm_shovel"]
            max_steps_shovel = python_data["max_steps_shovel"]
            steps_per_deg_turn = python_data["steps_per_deg_turn"]
            max_deg_turn = python_data["max_deg_turn"]
            lid_holder_slot = python_data["lid_holder_slot"]
            pipet_station_slot = python_data["pipet_station_slot"]
            measurement_slot = python_data["measurement_slot"]

    except Exception:
        log(msg = "json file could not be read")



from tkinter import *
from cytomat.utils import lazy_load_config_file
from cytomat.scripts.setup_cytomat import get_config_dir
from .manage_settings.manager_slots import ManagerSlots
from .manage_settings.manager_config_file import ManagerConfigFile
from sqlite3 import IntegrityError
from typing import  Optional
import tkinter.font

class ManagerSettings:
    __root: Toplevel
    def __init__(self, db, root, gui):
        self.__root = Toplevel(root)
        self.__root.title('Setting Manager')
        self.__root.geometry('1280x720')
        self.__root.withdraw()
        self.__db = db
        self.__gui: GUI = gui # type: ignore

        #from utils file
        self.__config_data_json: dict = lazy_load_config_file()
        #from setup_cytomat file
        self.__config_path = get_config_dir()

        self.sync_db_with_config_file()
        self.__root.protocol("WM_DELETE_WINDOW", self.on_close)
        self.check_box_var_foc = BooleanVar()
        self.__theme_pack: Optional[dict] = None
        self.__theme_pack_set_dark, self.__theme_pack_set_bright =  self.init_theme_pack()
        self.set_theme_pack('bright')

        self.__enter_label_settings:Optional[Label] = None
        self.__enter_label_gui: Optional[Label] = None
        self.__button_settings: Optional[Button] = None
        self.__button_gui: Optional[Button] = None
        self.__check_box_run_foc:Optional[Checkbutton] = None

        self.__default_hour_value: int = 3
        self.__hour_var = StringVar(value=str(self.__default_hour_value))
        self.__default_min_value = 0
        self.__minute_var = StringVar(value = str(self.__default_min_value))

        self.__label_foc_hour: Optional[Label] = None
        self.__entry_foc_hour: Optional[Entry] = None
        self.__entry_foc_minute: Optional[Entry] = None
        self.__label_foc_minute: Optional[Label] = None
        self.__label_middle: Optional[Label] = None
        self.__button_enter_foc: Optional[Button] = None

        self.check_box_var = BooleanVar()
        self.__check_box_theme: Optional[Checkbutton] = None

        self.top_level: Optional[ManagerSettingsTop] = None

    def get_root(self):
        return self.__root

    def display_window(self):
        self.__root.deiconify()

    def on_close(self):
        #self.no_button_callback()
        self.__root.withdraw()

    def create_overview(self):
        self.set_grid_settings()
        self.create_color_label()
        self.create_enter_label()
        self.create_button_gui()
        self.create_button_settings()
        self.create_theme_check_box()
        self.create_all_foc_check_box()
        self.create_entry_box_foc()
        self.create_enter_foc()

        #Ttheme pack is initialized in constructor
        self.aplay_theme_settings(self.__theme_pack)

    def set_grid_settings(self):
        self.__root.rowconfigure(0, weight = 1)
        self.__root.rowconfigure(1, weight = 1)
        self.__root.rowconfigure(2, weight = 1)
        self.__root.rowconfigure(3, weight = 1)
        self.__root.columnconfigure(0, weight = 4)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.columnconfigure(2, weight = 1)
        self.__root.columnconfigure(3, weight = 1)
        self.__root.columnconfigure(4, weight = 1)

    def create_enter_label(self):
        font = tkinter.font.Font(family = 'Times New Roman', size = 14)

        self.__enter_label_settings = Label(self.__root, text = 'Enter Settings', font=font)
        self.__enter_label_settings.grid(row=0, column=0)

        self.__enter_label_gui = Label(self.__root, text = 'Enter Cytomat GUI', font = font)
        self.__enter_label_gui.grid(row = 2, column=0)

    def create_button_settings(self):
        self.__button_settings = Button(self.__root, text = 'Enter', command = self.settings_button_callback)
        self.__button_settings.grid(row = 1, column=0, ipadx = 40, ipady = 10, sticky='n')

    def create_button_gui(self):
        self.__button_gui = Button(self.__root, text = 'Enter', command = self.gui_button_callback)
        self.__button_gui.grid(row = 3, column=0, ipadx = 40, ipady = 10, sticky='n')

    def settings_button_callback(self):
        self.create_top_level_window()

    def gui_button_callback(self):
        self.__root.withdraw()

    def create_all_foc_check_box(self):
        self.check_box_var_foc = BooleanVar()
        self.__check_box_run_foc = Checkbutton(self.__root, text = 'Run all foc at time', variable= self.check_box_var_foc, command = self.all_foc_check_box_callback)
        self.__check_box_run_foc.grid(row = 1, column=1, sticky='w')

    def all_foc_check_box_callback(self):
        if self.check_box_var_foc.get():
            self.__entry_foc_hour.config(state='normal', textvariable=self.__hour_var)
            self.__entry_foc_minute.config(state='normal', textvariable = self.__minute_var)
            self.__button_enter_foc.config(state='normal')
        else:
            text_var_dis = StringVar(value='')
            self.__entry_foc_hour.config(textvariable=text_var_dis, state='disabled')
            self.__entry_foc_minute.config(textvariable=text_var_dis, state='disabled')
            self.__button_enter_foc.config(state='disabled')
            self.__gui.set_run_all_foc_at_time(False)

    def create_entry_box_foc(self):
        self.__label_foc_hour = Label(self.__root,  text='hour')
        self.__label_foc_hour.grid(row=1, column=1, sticky='sw')

        self.__entry_foc_hour = Entry(self.__root, state='disabled')
        self.__entry_foc_hour.grid(row = 2, column = 1, sticky='nwe')

        self.__label_foc_minute = Label(self.__root, text='minute')
        self.__label_foc_minute.grid(row=1, column=3, sticky='sw')

        self.__entry_foc_minute = Entry(self.__root, state='disabled')
        self.__entry_foc_minute.grid(row = 2, column = 3, sticky='nwe')

        self.__label_middle = Label(self.__root, text = ':')
        self.__label_middle.grid(row = 2, column = 2, sticky='new', ipadx=10)

    def create_color_label(self):
        self.__color_label = Label(self.__root, bg = 'red')
        #TODO place label

    def create_enter_foc(self):
        self.__button_enter_foc = Button(self.__root, text = 'Enter ', command = self.enter_foc_button_callback, state='disabled')
        self.__button_enter_foc.grid(row = 1, column=3, rowspan=3, sticky = 'e')

    def enter_foc_button_callback(self):
        if not self.check_box_var_foc.get():
            return

        entry_min = self.__entry_foc_minute.get()
        entry_hour = self.__entry_foc_hour.get()

        if not self.is_entries_valid(entry_min=entry_min, entry_h=entry_hour):
            self.__entry_foc_hour.delete(0, END)
            self.__entry_foc_minute.delete(0, END)

            self.__hour_var.set(str(self.__default_hour_value))
            self.__minute_var.set(str(self.__default_min_value))
            return


        self.__hour_var.set(entry_hour)
        self.__minute_var.set(entry_min)
        entry_min = int(entry_min)
        entry_hour = int(entry_hour)
        self.__gui.set_run_all_foc_at_time(bool_run=True, timer_hour=entry_hour, timer_minute=entry_min)
        print(entry_hour, entry_min)

    def is_entries_valid(self, entry_min: str, entry_h: str):
        if not entry_min.isdigit() or not entry_h.isdigit():
            return

        entry_min = int(entry_min)
        entry_h = int(entry_h)

        if entry_h > 23 or entry_min>59:
            return

        return True

    def create_theme_check_box(self):
        self.check_box_var = BooleanVar()
        self.__check_box_theme = Checkbutton(self.__root, text = 'Dark Mode', variable= self.check_box_var, command = self.theme_check_box_callback)
        self.__check_box_theme.grid(row = 0, column=1, sticky='sw')

    def theme_check_box_callback(self):
        if self.check_box_var.get():
            self.set_theme_pack('dark')
        else:
            self.set_theme_pack('bright')

        self.aplay_theme_settings(self.__theme_pack)

    def set_theme_pack(self, theme):
        if theme == 'dark':
            self.__theme_pack = self.__theme_pack_set_dark

        if theme == 'bright':
            self.__theme_pack =  self.__theme_pack_set_bright


    def get_theme_pack(self) -> dict:
        return self.__theme_pack

    def init_theme_pack(self)->tuple[dict, dict]:
        theme_pack_dark: dict = {'family': 'dark', 'entry': '#616161', 'button': '#424242', 'fg': 'white',
                                 'bg': '#454545', 'header': '#707070', 'check_box': '#626969', 'label': '#a2a6a4','fg_check_box': 'grey'}

        theme_pack_bright: dict = {'family': 'bright', 'entry': 'lightgrey', 'button': 'white', 'fg': 'black',
                                   'bg': '#E5E5E5', 'header': '#BDBDBD', 'check_box': '#E5E5E5', 'label': '#dad6cb', 'fg_check_box': 'white'}

        return theme_pack_dark, theme_pack_bright


    def aplay_theme_settings(self, theme_pack: dict):
        """
        configures the colors of the widgets according to the theme pack dictionary

        """
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        check_box = theme_pack['check_box']
        fg_check_box = theme_pack['fg_check_box']
        label = theme_pack['label']

        self.__root.config(bg=bg)
        self.__button_settings.config(bg = button, fg = fg)
        self.__button_gui.config(bg = button, fg = fg)
        self.__enter_label_settings.config(fg = fg, bg = bg)
        self.__enter_label_gui.config(bg = bg, fg = fg)
        self.__check_box_theme.config(bg = check_box, fg = fg, selectcolor = fg_check_box)
        self.__check_box_run_foc.config(bg = check_box, fg =  fg, selectcolor=fg_check_box)
        self.__label_foc_hour.config(bg = bg, fg = fg)
        self.__label_foc_minute.config(bg = bg, fg = fg)
        self.__button_enter_foc.config(bg = button, fg  = fg)
        self.__label_middle.config(bg = bg, fg = fg)


        try:
            self.top_level.applay_theme_pack(theme_pack=self.__theme_pack)
        except Exception:
            pass

    def create_top_level_window(self):
        self.top_level = ManagerSettingsTop(root = self.__root, config_data=self.__config_data_json, config_path=self.__config_path, db = self.__db)
        self.top_level.create_overview()
        self.top_level.applay_theme_pack(theme_pack=self.__theme_pack)

    def sync_db_with_config_file(self)->None:
        """
        updates the Database after the json file has been changed
        """
        for key, value in self.__config_data_json.items():
            if not 'slot' in key:
                continue

            if not value:
                continue

            try:
                self.__db.insert_new_row(slot = value, role = 1)

            except IntegrityError:
                continue
        #Create Rows to Setup Slots
    def submit_callback(self):
        pass

class ManagerSettingsTop:
    __root: Toplevel
    def __init__(self, root, db, config_path, config_data):
        self.__root = Toplevel(root)
        self.__root.title('Settings')
        self.__root.geometry('1920x1080')
        
        self.__db = db
        self.__config_data_json = config_data
        self.__config_path = config_path

        self.__right_frame: Optional[Frame] = None
        self.__left_frame: Optional[Frame] = None

        self.__manager_config: Optional[ManagerConfigFile] = None
        self.__manager_slots: Optional[ManagerSlots] = None

        self.__theme_pack: Optional[dict] = None
        self.setup_frames()

    def get_root(self):
        return self.__root
        
    def create_overview(self):
        self.__manager_config = ManagerConfigFile(root = self.__left_frame, data = self.__config_data_json, config_directory_path = self.__config_path, db=self.__db)
        self.__manager_slots = ManagerSlots(root = self.__right_frame, db = self.__db)
        
        self.__manager_config.create_overview()
        self.__manager_slots.create_overview()

    def setup_frames(self):
        self.__left_frame = Frame(self.__root)
        self.__left_frame.grid(row = 0, column = 0, sticky = 'nsew')
                
        self.__right_frame = Frame(self.__root, bg = 'lightblue')
        self.__right_frame.grid(row = 0, column = 1, sticky = 'nsew')
        
        self.__root.columnconfigure(0, weight=1)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.rowconfigure(0, weight = 1)
    
    def applay_theme_pack(self, theme_pack):
        self.__theme_pack = theme_pack
        self.__manager_config.aplay_theme_pack(theme_pack = self.__theme_pack)
        self.__manager_slots.apply_theme_pack(theme_pack = self.__theme_pack)

class ThemePack:
    def __init__(self, root):
        self.__root = root